/*
 * This file is part of the ALT project, module Alt-App-Starter.
 * Copyright (c) BaseALT Ltd 2019,2020 Pavel Moseev <mars@altlinux.org>
 *
 * This is free software; you can use this application under the GNU
 * General Public License, version 2. See the file "COPYING" for the
 * exact licensing terms.
 */

#include "mainwindow.h"

#include "ui_mainwindow.h"

#include <KDESu/Client>
#include <kdesu/defaults.h>
#include <kdesu/suprocess.h>
#include <KOpenWithDialog>
#include <kstartupinfo.h>
#include <KUser>
#include <KCoreAddons/kshell.h>

#include <QByteArray>
#include <QCompleter>
#include <QDebug>
#include <QDir>
#include <QFile>
#include <QFileDialog>
#include <QLocale>
#include <QMessageBox>
#include <QSettings>
#include <QStandardPaths>
#include <QTextStream>

#include <boost/scope_exit.hpp>

#include <pwd.h>
#include <sys/resource.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <poll.h>
#include <sys/wait.h>

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    QFont font = ui->label_consoleNotification->font();
    int fontSize = font.pointSize();
    font.setPointSize(fontSize-1);
    ui->label_consoleNotification->setFont(font);

    auto checkBox_runInTerminal_lambda = [this](bool checked)
    {
        if (checked)
        {
            switch (getTerminalType())
            {
            case terminal::konsole:
            case terminal::xterm:
                ui->checkBox_dontCloseTerminal->setEnabled(true);
                break;

            default:
                ui->checkBox_dontCloseTerminal->setEnabled(false);
                break;
            }
        }
        else
        {
            ui->checkBox_dontCloseTerminal->setEnabled(false);
        }
    };

    auto checkBox_changePriority_lambda = [this](bool checked)
    {
        this->ui->prioritySlider->setEnabled(checked);
        this->ui->prioritySpinBox->setEnabled(checked);
        this->ui->label_priority->setEnabled(checked);
        this->ui->label_highPrior->setEnabled(checked);
        this->ui->label_lowPrior->setEnabled(checked);
        this->checkUserPermissions();
    };

    auto checkBox_changeUser_lambda = [this](bool checked)
    {
        this->ui->lineEdit_userPassword->setEnabled(checked);
        this->ui->label_userName->setEnabled(checked);
        this->ui->label_userPassword->setEnabled(checked);
        this->ui->comboBox_usersList->setEnabled(checked);
        this->checkUserPermissions();
    };

    checkBox_runInTerminal_lambda(ui->checkBox_runInTerminal->isChecked());
    checkBox_changePriority_lambda(ui->checkBox_changePriority->isChecked());
    checkBox_changeUser_lambda(ui->checkBox_changeUser->isChecked());

    // Read all users names from KUser and fill comboBox_usersList
    {
        const unsigned int defminuid = 500;

        for (const auto &kuser: KUser::allUsers())
        {
            if ((kuser.userId().nativeId() >= defminuid) || (kuser.userId().nativeId() == 0))
            {
                ui->comboBox_usersList->addItem(kuser.loginName());
            }
        }
    }

    // Fill ComboBox with history of user commands:
    m_commandListFromFile = readCommandHistory();
    ui->comboBox_commandInput->addItems(m_commandListFromFile); // Save command history to ComboBox
    QCompleter *completer = new QCompleter(m_commandListFromFile, this);
    completer->setCaseSensitivity(Qt::CaseInsensitive);
    ui->comboBox_commandInput->setCompleter(completer);
    ui->comboBox_commandInput->setCurrentIndex(-1);

    connect(ui->pushButton_enterForm, &QPushButton::clicked, this, &MainWindow::runApplication);
    connect(ui->pushButton_closeForm, &QPushButton::clicked, this, &MainWindow::close);
    connect(ui->pushBtn_chooseAppFromMenu, &QPushButton::clicked, this, &MainWindow::chooseApplicationFromMenu);
    connect(ui->pushBtn_chooseAppFromBin, &QPushButton::clicked, this, &MainWindow::chooseApplicationFromBin);

    connect(ui->checkBox_runInTerminal, &QCheckBox::clicked, this, checkBox_runInTerminal_lambda);
    connect(ui->checkBox_changePriority, &QCheckBox::clicked, this, checkBox_changePriority_lambda);
    connect(ui->checkBox_changeUser, &QCheckBox::clicked, this, checkBox_changeUser_lambda);

    connect(ui->prioritySlider, &QSlider::valueChanged, this, [this](int value) {
        this->ui->prioritySpinBox->setValue(value);
        this->checkUserPermissions();
    });

    connect(ui->prioritySpinBox, QOverload<int>::of(&QSpinBox::valueChanged), this, [this](int value) {
        this->ui->prioritySlider->setValue(value);
        this->checkUserPermissions();
    });
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::setAppName(const QString &chosenAppName) // Parse the name of chosen application in main menu
{
    if (!chosenAppName.isEmpty())
    {
        QString inputString = chosenAppName; // Input param, example: altappstart://xterm

        if (inputString.startsWith(QStringLiteral("altappstart://"), Qt::CaseInsensitive))
        {
             inputString = inputString.remove(0, QStringLiteral("altappstart://").length());
        }

        KService::Ptr service = KService::serviceByDesktopName(inputString);
        if (service)
        {
            processAppType(service); // Check is it terminal application or not.
        }
    }
}

void MainWindow::checkUserPermissions() // If not root and need to change process priority
{
    if ((!ui->checkBox_changeUser->isChecked()) && ui->checkBox_changePriority->isChecked())
    {
        int rlim_max = readCurrentUserLimit(); // Read current user rights

        if (ui->prioritySpinBox->value() <= rlim_max)
        {
            ui->prioritySpinBox->setValue(rlim_max);
            ui->prioritySlider->setValue(rlim_max);
        }
    }
}

QStringList MainWindow::readCommandHistory() // Read command history from file
{
    QFile historyFile(QStandardPaths::writableLocation(QStandardPaths::AppDataLocation) + QDir::separator() + QStringLiteral("command-history"));

    QStringList result;

    if ((historyFile.exists()) && (historyFile.open(QIODevice::ReadOnly)))
    {
        QString str;

        while (!historyFile.atEnd())
        {
            str = historyFile.readLine();
            str = str.trimmed();
            result << str;
        }

        historyFile.close();
    }

    return result;
}

void MainWindow::saveCommandHistory(const QString &command) // Save command history to the file
{
    const QString filePath = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);

    QDir historyDir(filePath);
    if (!historyDir.exists())
    {
        qInfo() << "Command history directory doesn't exist! Create it.";
        QDir().mkpath(filePath);
    }

    QFile historyFile(filePath + QDir::separator() + QStringLiteral("command-history"));

    int commandIndex = m_commandListFromFile.indexOf(command, 0);
    if (commandIndex == -1)
    {
        if (m_commandListFromFile.count() < m_commandHistoryLimit)
        {
            if (historyFile.open(QIODevice::Append | QIODevice::Text))
            {
                QTextStream writeStream(&historyFile);
                writeStream << command + QStringLiteral("\n");
                historyFile.close();
            }
        }
        else
        {
            m_commandListFromFile.removeAt(0); // Delete first string from List of commands
            if (historyFile.open(QIODevice::WriteOnly | QIODevice::Text))
            {
                QTextStream writeStream(&historyFile);

                for (int i = 0; i < m_commandListFromFile.size(); ++i)
                {
                    writeStream << m_commandListFromFile.at(i) << '\n';
                }

                writeStream << command + QStringLiteral("\n");
                historyFile.close();
            }
        }
    }
}

void MainWindow::chooseApplicationFromMenu() // Choose app name from list of menu items
{
    KOpenWithDialog dialog(this);
    dialog.setWindowModality(Qt::WindowModal);
    if (dialog.exec())
    {
        KService::Ptr service = dialog.service();
        if (service)
        {
            processAppType(service); // Check is it terminal application or not.
        }
    }
}

void MainWindow::chooseApplicationFromBin() // Choose app name from bin directory
{
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"), "/usr/bin/", tr("All files (*.*)"));
    if (!fileName.isEmpty())
    {
        ui->comboBox_commandInput->setCurrentText(fileName);
    }
}

int MainWindow::readCurrentUserLimit() // Read current user rights
{
    struct rlimit nice_limits;
    nice_limits.rlim_cur = 0;
    nice_limits.rlim_max = 0;

    getrlimit(RLIMIT_NICE, &nice_limits);
    int rlim_max = nice_limits.rlim_max;

    return rlim_max;
}

void MainWindow::runApplication() // Run command with or without user changing
{
    QString fullCommandName = ui->comboBox_commandInput->currentText().simplified(); // All params, Ex:  "libreoffice6.3 --draw %U"
    bool useAnotherUser = ui->checkBox_changeUser->isChecked();
    bool runInTerminal = ui->checkBox_runInTerminal->isChecked();
    bool changePriority = ui->checkBox_changePriority->isChecked();

    int prio = ui->prioritySpinBox->value();

    if (fullCommandName.isEmpty())
    {
        QMessageBox::warning(this, tr("Application call error"), tr("Enter the name of the application!\nFor example: 'xterm' or 'konsole'."));
        return;
    }

    QStringList command_array = KShell::splitArgs(fullCommandName);
    if (command_array.isEmpty())
    {
        QMessageBox::warning(this, tr("Application name error"), tr("Enter the name of the application!\nFor example: 'xterm' or 'konsole'."));
        return;
    }

    if (QStandardPaths::findExecutable(command_array.first()).isEmpty()) // Check entered application name
    {
        QMessageBox::warning(this, tr("Application name error"), tr("The entered application name could not be found."));
        return;
    }

    if (runInTerminal) // Open app in terminal emulator (it choosen terminal app from system settings).
    {
        command_array = formTerminalCommand(command_array);
        if (command_array.isEmpty())
        {
            QMessageBox::warning(this, tr("Open terminal application error"), tr("Terminal emulator not found."));
            return;
        }
    }

    if (useAnotherUser) // Run command with user changing.
    {
        QString userName = ui->comboBox_usersList->currentText().simplified();

        if (userName.isEmpty())
        {
            QMessageBox::warning(this, tr("User name error"), tr("User name not entered.\nPlease enter user name."));
            return;
        }

        KDESu::KDEsuClient client;
        KDESu::SuProcess proc;

        QString password = ui->lineEdit_userPassword->text();

        QList<QByteArray> env;
        env << (QByteArray("DESKTOP_STARTUP_ID=") + KStartupInfo::startupId());
        KUser u;
        env << (QStringLiteral("KDESU_USER=") + u.loginName()).toLocal8Bit();

        if (!client.isServerSGID()) // Check for daemon and start it if necessary:
        {
            QMessageBox::warning(this, tr("Daemon error"), tr("Daemon not safe (not sgid), not using it."));
            qDebug() << "Alt-App-Starter message. Daemon not safe (not sgid), not using it.\n";
        }
        else if (client.ping() == -1)
        {
            if (client.startServer() == -1) // Try to start daemon
            {
                QMessageBox::warning(this, tr("Daemon start error"), tr("Could not start daemon, functionality is reduced."));
                qDebug() << "Alt-App-Starter message. Could not start daemon, functionality is reduced.\n";
            }

            qDebug() << "Alt-App-Starter message. Daemon just started.\n";
        }

        struct passwd *pw = getpwnam(userName.toLocal8Bit().constData()); // User verification
        if (pw == nullptr)
        {
            QMessageBox::warning(this, tr("User name error"), tr("User %1 does not exist!").arg(userName));
            qDebug() << "User " << userName << " does not exist\n";
            return;
        }

        proc.setUser(userName.toLocal8Bit());
        int needPass = proc.checkNeedPassword(); // Checks if a password is needed.
        int passStatus = proc.checkInstall(password.toLocal8Bit().constData()); // Password verification

        if (needPass < 0)
        {
            qDebug() << "Su returned with an error.\n";
        }

        if (needPass == 0)
        {
            qDebug() << "Don't need password!\n";
        }

        if (passStatus != 0) // Returns: Zero on success
        {
            if (needPass == 0)
            {
                QMessageBox::warning(this, tr("Password error"), tr("Password not entered.\nPlease enter user password."));
            }
            else
            {
                QMessageBox::warning(this, tr("Password error"), tr("Wrong password or user privileges!"));
            }

            return;
        }

        client.setPass(password.toLocal8Bit().constData(), KDESu::defTimeout);
        if (changePriority && useAnotherUser)  // Change process priority
        {
            int rlim_max = readCurrentUserLimit();
            KUser kuser(userName);
            if ((kuser.userId().nativeId() == 0) || ((kuser.userId().nativeId() != 0) && (prio >= rlim_max)))
            { // root user or another user with priority [rlim, 19]
                client.setPriority((prio*(-1)+20)*2.5);
            }
        }

        this->hide();
        int result = client.exec(KShell::joinArgs(command_array).toLocal8Bit(), userName.toLocal8Bit(), QByteArray(), env);

        BOOST_SCOPE_EXIT(&client)
        {
            if (client.ping() == -1) // Daemon stop
            {
                qDebug() << "Alt-App-Starter message. Daemon not running - nothing to stop.\n";
            }

            if (client.stopServer() != -1)
            {
                qDebug() << "Alt-App-Starter message. Daemon is running. Stop it.\n";
            }
        } BOOST_SCOPE_EXIT_END;

        if (result != 0) // Returns: Zero on success, -1 on failure.
        {
            this->show();
            QMessageBox::warning(this, tr("Application call error"), tr("Application launch failed!"));
            return;
        }

        client.exitCode();
    }
    else // Run command without user changing
    {
        // prepare arguments for exec
        QList<QByteArray> args_list_lifetimes;
        QVector<char*> args_list_processed;

        for (const auto &item: command_array)
        {
            args_list_lifetimes.push_back(item.toLocal8Bit());
        }

        for (auto &item: args_list_lifetimes)
        {
            args_list_processed.push_back(item.data());
        }

        args_list_processed.push_back(nullptr);

        int pipes[2] = { -1, -1 };

        if (pipe(pipes) < 0)
        {
            QMessageBox::warning(this, tr("Application call error"), tr("%1 call failed with error code %2: %3").arg(QStringLiteral("pipe()")).arg(errno).arg(strerror(errno)));
            return;
        }

        BOOST_SCOPE_EXIT(&pipes)
        {
            if (pipes[0] != -1)
            {
                ::close(pipes[0]);
            }

            if (pipes[1] != -1)
            {
                ::close(pipes[1]);
            }
        } BOOST_SCOPE_EXIT_END;

        if (fcntl(pipes[1], F_SETFD, FD_CLOEXEC) < 0)
        {
            QMessageBox::warning(this, tr("Application call error"), tr("%1 call failed with error code %2: %3").arg(QStringLiteral("fcntl()")).arg(errno).arg(strerror(errno)));
            return;
        }

        this->hide();

        int fork_result = fork();
        if (fork_result < 0)
        {
            this->show();
            QMessageBox::warning(this, tr("Application call error"), tr("%1 call failed with error code %2: %3").arg(QStringLiteral("fork()")).arg(errno).arg(strerror(errno)));
            return;
        }

        if (fork_result == 0)
        {
            // child
            ::close(pipes[0]);
            pipes[0] = -1;

            if (changePriority) // Try to change process priority
            {
                // if setpriority() failed, consider it bad enough to abort running and report error
                if (setpriority(PRIO_PROCESS, getpid(), prio) < 0)
                {
                    dprintf(pipes[1], "%s\n", tr("%1 call failed with error code %2: %3").arg(QStringLiteral("setpriority()")).arg(errno).arg(strerror(errno)).toLocal8Bit().data());
                    _exit(EXIT_FAILURE);
                }
            }

            execvp(args_list_processed[0], args_list_processed.data());
            dprintf(pipes[1], "%s\n", tr("%1 call failed with error code %2: %3").arg(QStringLiteral("execvp()")).arg(errno).arg(strerror(errno)).toLocal8Bit().data());
            _exit(EXIT_FAILURE);
        }

        bool reap_child = true;

        BOOST_SCOPE_EXIT(&reap_child, &fork_result)
        {
            if (reap_child)
            {
                for (;;)
                {
                    int dummy = 0;
                    auto child_pid = waitpid(fork_result, &dummy, 0);
                    if (child_pid == fork_result)
                    {
                        break;
                    }
                }
            }
        } BOOST_SCOPE_EXIT_END;

        // parent
        ::close(pipes[1]);
        pipes[1] = -1;

        QString child_output;
        pollfd pollfds;

        for (;;)
        {
            pollfds.fd = pipes[0];
            pollfds.events = POLLIN | POLLHUP;
            pollfds.revents = 0;

            int result = poll(&pollfds, 1, -1);
            if (result < 0)
            {
                this->show();
                QMessageBox::warning(this, tr("Application call error"), tr("Failed to get launch result of child: %1").arg(tr("%1 call failed with error code %2: %3").arg(QStringLiteral("poll()")).arg(errno).arg(strerror(errno))));
                return;
            }

            if (pollfds.revents & POLLIN)
            {
                char buffer[4096];
                auto read_size = read(pipes[0], buffer, sizeof(buffer));
                if (read_size < 0)
                {
                    this->show();
                    QMessageBox::warning(this, tr("Application call error"), tr("Failed to get launch result of child: %1").arg(tr("%1 call failed with error code %2: %3").arg(QStringLiteral("read()")).arg(errno).arg(strerror(errno))));
                    return;
                }

                child_output += QString::fromLocal8Bit(buffer, read_size);
            }
            else if (pollfds.revents & POLLHUP)
            {
                break;
            }
        }

        if (!child_output.isEmpty())
        {
            this->show();
            QMessageBox::warning(this, tr("Application call error"), child_output);
            return;
        }

        reap_child = false;
    }

    saveCommandHistory(ui->comboBox_commandInput->currentText());
    this->close();
}

QString MainWindow::processPlaceholders(const KService::Ptr &service) // Checking the input parameters required to run the application.
{
    QString serviceExecName = service->exec().simplified();
    if (serviceExecName.isEmpty())
    {
        return QString();
    }

    QStringList allInputParams = KShell::splitArgs(serviceExecName);
    QStringList outputParams;

    for (auto iter = allInputParams.begin(); iter != allInputParams.end(); ++iter)
    {
        if ((iter->isEmpty()) || ((*iter)[0] != '%'))
        {
            outputParams.push_back(*iter);
        }
        else
        {
            if (*iter == QStringLiteral("%i"))
            {
                QString iconNameInDesktopFile = service->icon(); // Find icon name in desktop file.
                if (iconNameInDesktopFile.isEmpty())
                {
                    QMessageBox::warning(this, tr("Desktop file parsing error"),
                                               tr("Desktop icon placeholder \"%%i\" is used but \"Icon\" entry is missing"));
                    return QString();
                }

                outputParams.push_back(QStringLiteral("--icon"));
                outputParams.push_back(iconNameInDesktopFile);
            }
            else if (*iter == QStringLiteral("%c"))
            {
                outputParams.push_back(service->name()); // Find name in desktop file.
            }
            else if (*iter == QStringLiteral("%k"))
            {
                outputParams.push_back(service->entryPath()); // Find path to desktop file.
            }
            else if (*iter == QStringLiteral("%%"))
            {
                outputParams.push_back(QStringLiteral("%"));
            }
            else
            {
                // skip all listed placeholders. If unknown placeholder is encountered, return error.
                const QStringList placeholdersList = { "%f", "%F", "%u", "%U", "%d", "%D", "%n", "%N", "%v", "%m" };

                if (!placeholdersList.contains(*iter))
                {
                    QMessageBox::warning(this, tr("Desktop file parsing error"),
                                               tr("Unknown placeholder \"%1\" is encountered").arg(*iter));
                    return QString();
                }
            }
        }
    }

    return KShell::joinArgs(outputParams);
}

MainWindow::terminal MainWindow::getTerminalType()
{
    const QMap<QString, terminal> terminal_map = {
        { "konsole", terminal::konsole },
        { "xterm", terminal::xterm },
        { "xvt", terminal::xvt },
    };

    QSettings settings(QStandardPaths::writableLocation(QStandardPaths::ConfigLocation) + QDir::separator() + QStringLiteral("kdeglobals"), QSettings::IniFormat);

    QString exec_termApp = settings.value("General/TerminalApplication", "konsole").toString(); // Read TerminalApplication settings from "kdeglobals".

    {
        int last_index = exec_termApp.lastIndexOf(QDir::separator());
        if (last_index != -1)
        {
            exec_termApp = exec_termApp.mid(last_index + 1);
        }
    }

    if (exec_termApp.isEmpty())
    {
        return terminal::unsupported;
    }

    // first try selected terminal
    {
        auto iter = terminal_map.find(exec_termApp);
        if (iter != terminal_map.end())
        {
            if (!QStandardPaths::findExecutable(iter.key()).isEmpty())
            {
                return iter.value();
            }
        }
    }

    // after that try every supported terminal
    {
        for (auto iter = terminal_map.begin(); iter != terminal_map.end(); ++iter)
        {
            if (!QStandardPaths::findExecutable(iter.key()).isEmpty())
            {
                return iter.value();
            }
        }
    }

    return terminal::unsupported;
}

QStringList MainWindow::formTerminalCommand(const QStringList &command)
{
    bool dontCloseTerminal = (ui->checkBox_dontCloseTerminal->isEnabled() && ui->checkBox_dontCloseTerminal->isChecked());

    switch (getTerminalType())
    {
    case terminal::konsole:
        return (dontCloseTerminal ? QStringList { "kde5", "konsole", "--noclose", "-e" } : QStringList { "kde5", "konsole", "-e" }) + command;
        break;

    case terminal::xterm:
        return (dontCloseTerminal ? QStringList { "xterm", "-hold", "-e" } : QStringList { "xterm", "-e" }) + command;
        break;

    case terminal::xvt:
        return QStringList { "xvt", "-e" } + command;
        break;

    case terminal::unsupported:
    default:
        break;
    }

    return QStringList();
}

void MainWindow::processAppType(const KService::Ptr &service) // Use data about app type to update info in user interface.
{
    ui->comboBox_commandInput->setCurrentText(processPlaceholders(service));

    bool isTerm = service->terminal();
    if (isTerm)
    {
        ui->checkBox_runInTerminal->setText(tr("Run in terminal* (Recommended for %1)").arg(service->name()));
    }
    else
    {
        ui->checkBox_runInTerminal->setText(tr("Run in terminal*"));
    }

    ui->checkBox_runInTerminal->setChecked(isTerm);
}
